%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for HP-UX
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 09-Oct-87 (Leigh Stoller)
%  Do not call build-trap-message anymore. It is too flakey to be relied
%   on to tell the truth. 
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is received.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn
   (setf *sigcalls* nil)
   %
   % Create a list of initialization code needed to set up the traps.
   % The return value is the entry points for each signal handler.
   % This is later flagged as a cmacro so the returned code can be
   % inserted directly into the lap coded procedure
   % 'initializeinterrupts each time it is called.  The list
   % *sigentries is then inserted to define the entry points.
   (de *sigsetup (signumber function handler errorstring)
       (setf *sigcalls* (append 
			 % Sigset takes a function pointer.
			 `((*move (wconst ,signumber) (reg 1))
			   (*move ,handler (reg 2))
			   (*link psl_sigset expr 2))
			 *sigcalls*))
       % Return the the function definition for the signal handler.
       `((*entry ,function expr 0)
	 ,handler
	 (*move (wconst ,signumber) (reg 1))
	 (*move ,handler (reg 2))
	 (*link sigrelse expr 2)
	 (*move (fluid nil) (reg nil))
	 (*move (quote ,errorstring) (reg 1))
	 (*jcall sigunwind))
       )
   
   % Return the entry point list. Defined as a cmacro.
   (de *sigcall ()
       *sigcalls*)

   (defcmacro *sigcall)
   (defcmacro *sigsetup)

   )
) % End of compiletime.


(lap '(
   % (*sigsetup 1  Huphandler  Huphandlerinstruction  "Hup")
       (*sigsetup 2  Inthandler  IntHandlerInstruction  "Interrupt")
       (*sigsetup 3  QuitHandler QuitHandlerInstruction "Quit")
       (*sigsetup 4  IllHandler  IllHandlerInstruction  "Illegal Instruction")
       (*sigsetup 5  Traphandler TrapHandlerInstruction "Trace Trap")
       (*sigsetup 6  IotHandler  IotHandlerInstruction  "IOT Instruction")
       (*sigsetup 7  Emthandler  EmtHandlerInstruction  "EMT Instruction")
       (*sigsetup 8  FpeHandler  FpeHandlerInstruction "Floating Pt Exception")
       (*sigsetup 10 Bushandler  BusHandlerInstruction  "Bus Error")
       (*sigsetup 11 SegHandler  SegHandlerInstruction  
		                            "Segmentation Violation")
       (*sigsetup 12 SysHandler  SysHandlerInstruction  
		                            "Bad Args to System Call")
       (*sigsetup 13 PipeHandler PipeHandlerinstruction 
		                            "Write on Pipe With Noone to Read")
       (*sigsetup 14 AlrmHandler AlrmHandlerInstruction "Alarm Clock")
       (*sigsetup 19 PwrHandler  PwrHandlerInstruction  "Power-Fail Restart")

       (*entry initializeinterrupts expr 0)
       (*sigcall)
       (*exit 0)))


(lap 
 '((!*entry sigunwind expr 0)
     % At this point, an arg is already set up in register 1 with a message
     % describing the kind of trap.  First, though, we want to find the 
     % name of the routine that caused the trap by using the "interrupted
     % address" in the signal handler stack frame.  Then we want to
     % pop this frame off the stack.  It adds up to 476 bytes.

     % (*move (memory (reg st) 94) (reg 2))  % Addr. at which signal occurred
     (*wplus2 (reg st) 476)                % Pop signal-handling frame
     % (*link build-trap-message expr 2)     % This leaves its result in reg
                                           % 1, so the new message is
     (*jcall stderror)                     % ready to be passed
     ))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (if (funboundp 'code-address-to-symbol)
        (setf extra-info 
	      (bldmsg "%w%n%w%n%w"
		      " : the name of the routine that trapped can't be"
		      " reported unless the function CODE-ADDRESS-TO-SYMBOL"
		      " has been defined, by loading ADDR2ID."))
	% else, get the name of the offending function 
	(setf extra-info (bldmsg "%w%w" 
				 " in " 
				 (code-address-to-symbol (inf trap-addr))))
      )
      (bldmsg "%w%w" trap-type extra-info)))

% End of file.

